// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997 - 2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	Bivariate2DInterpolantRectangular.cc --- Implementation of
//											Bivariate2DInterpolantRectangular
//
//	DESCRIPTION:
//		Implementation of the Bivariate2DInterpolantRectangular class via the
//		TOMS routine QSHEP2. cfortran.h is used to interface to TOMS.
//
//	CHANGE HISTORY:
//		UL, 2000-01-11: 1.0.0: - first version
//		UL, 2000-01-25: 1.1.0: - changed from TOMS526 to TOMS660
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/slatec/cxx/Bivariate2DInterpolantRectangular.cc,v 1.3 2009/02/06 17:51:41 eojero Exp $
//

#include "slatec++.h"
#include "slatec.h"
#include "errstr.h"
#include "Tracer.h"

#include "Bivariate2DInterpolantRectangular.h"

namespace Slatec {

template <> Bivariate2DInterpolantRectangular<float>::Bivariate2DInterpolantRectangular(
											const std::vector<float> &x,
											const std::vector<float> &y,
											const std::vector<float> &z)
{
	PROC

	if (x.size() < 2)
		errstr<<"Need at least two X values to define bi-variate "
				"interpolator"<<Sas::error("TooLittleDataPoints");
	if (y.size() < 2)
		errstr<<"Need at least two Y values to define bi-variate "
				"interpolator"<<Sas::error("TooLittleDataPoints");

	_nxd	= x.size();
	_nyd	= y.size();
	_xd		= new float[_nxd];
	_yd		= new float[_nyd];
	_zd		= new float[_nxd*_nyd];
	_wk		= new float[3*_nxd*_nyd];

	for (int n=0; n<_nxd; ++n)
		_xd[n] = x[n];
	for (int n=0; n<_nyd; ++n)
		_yd[n] = y[n];
	for (int n=0; n<_nxd*_nyd; ++n)
		_zd[n] = z[n];

	int ierr = 0;
	int md = 1;	//	first call with MD=1 to set things up
	float xp = _xd[0],
		  yp = _yd[0],
		  zv;
	TOMS_RGBI3P(md, _nxd, _nyd, _xd, _yd, _zd, 1, &xp, &yp, &zv, ierr, _wk);

	if (ierr) {
		errstr<<"In setting up 2D-interpolation routine (TOMS760): ";
		switch (ierr) {
			case 1:	errstr<<"NXD is 1 or less";
					break;
			case 2: errstr<<"NYD is 1 or less";
					break;
			case 3: errstr<<"identical XD values or XD values out of sequence";
					break;	
			case 4: errstr<<"identical YD values or YD values out of sequence";
					break;
			case 5: errstr<<"NIP = 0 or less";
			default:
					errstr<<"an unknown error "<<ierr<<" occured.";
					break;
		}
		errstr<<Sas::error("NumericalError");
	}

	ENDPROC
} // Bivariate2DInterpolantRectangular::Bivariate2DInterpolantRectangular()


template <> float Bivariate2DInterpolantRectangular<float>::operator()(float x, float y) const
{
	PROC

	int md = 2, ierr = 0;
	float z;
	TOMS_RGBI3P(md, _nxd, _nyd, _xd, _yd, _zd, 1, &x, &y, &z, ierr, _wk);
	if (ierr)
		errstr<<"Unknown error "<<ierr<<" occurred in 2D-interpolation "
				"routine (TOMS760)"
			  <<Sas::error("NumericalError");

	RETURN z;
} // Bivariate2DInterpolantRectangular<float>::operator()


template <> Bivariate2DInterpolantRectangular<float>::~Bivariate2DInterpolantRectangular()
{
	PROC

	delete [] _xd; _xd = 0;
	delete [] _yd; _yd = 0;
	delete [] _zd; _zd = 0;
	delete [] _wk; _wk = 0;

	ENDPROC
} // Bivariate2DInterpolantRectangular<double>::~Bivariate2DInterpolantRectangular()

}

