// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
/* written by John Doug Reynolds, January 2000 */

#include "Convolver.h"
#include <slatec.h>
#include <errstr.h>

Convolver::Convolver( unsigned size )
   : _length(0)
{
   resize( size );
}

void	Convolver::resize( unsigned size )
{
   if ( size != _length ) {
      _length = size;

      _signal.resize( _length );
      _WA.resize( _length );
      _CH.resize( _length );

      SLATEC_RFFTI1( _length, &_WA.front(), _IFAC );
   }
   reset();
}

void	Convolver::reset( enum Type which )
{
   if ( which & SIGNAL ) _isOpen = false;
   if ( which & KERNEL ) _kernel.clear();
}

void	Convolver::testSize( const std::vector<float> &v ) const
{
   if ( v.size() != _length )
      errstr << "Attempt to load vector size " << v.size()
	     << " after initialization to size " << _length
	     << Sas::error("Convolver:sizeError");
}

void	Convolver::load( const std::vector<float> &v, enum Type to )
{
   testSize(v);
   load( &v.front(), to );
}

void	Convolver::load( std::vector<float> &v, enum Type to )
{
   testSize(v);
   load( &v.front(), to );
}

void Convolver::loadAsConst( std::vector<float> &v, enum Type to )
{
   testSize(v);
   load( static_cast<const float*>(&v.front()), to );
}

void Convolver::loadAsConst( float *v, enum Type to )
{
   load( static_cast<const float*>(v), to );
}

void	Convolver::combine( float *dst, const float *src )
{
   if ( _length == 0 ) return;

   unsigned count = (_length - 1) / 2;

   *dst++ *= *src++;

   while ( count-- ) {
      float srcR = *src++;
      float srcI = *src++;

      float dstR = dst[0];
      float dstI = dst[1];

      *dst++ = dstR * srcR - dstI * srcI;
      *dst++ = dstR * srcI + dstI * srcR;
   }

   if ( (_length & 1) == 0 ) *dst *= *src;
}

void	Convolver::load( const float *v, enum Type to )
{
   std::vector<float> temp( v, v + _length );
   load( &temp.front(), to );
}

void	Convolver::load( float *src, enum Type to )
{
   SLATEC_RFFTF1( _length, src, &_CH.front(), &_WA.front(), _IFAC );

   switch ( to ) {
   case KERNEL:
      if ( _kernel.empty() ) {
	 _kernel.resize( _length );
	 float *dst = &_kernel.front();
	 for ( unsigned count = _length ; count-- ; *dst++ = *src++ ) ;
      }
      else combine( &_kernel.front(), src );
      break;

   case SIGNAL:
      if ( ! _isOpen ) {
	 float *dst = &_signal.front(), norm = 1.0 / _length;
	 for ( unsigned count = _length ; count-- ; *dst++ = *src++ * norm ) ;
	 _isOpen = true;
      }
      else combine( &_signal.front(), src );
      break;

   default:
      errstr << "load(," << int(to) << "): target must be "
	     << int(SIGNAL) << " (SIGNAL) or " << int(KERNEL) << " (KERNEL)"
	     << Sas::error("Convolver:invalid");
   }
}

const std::vector<float>&	Convolver::operator() (void)
{
   if ( _isOpen ) {
      if ( ! _kernel.empty() ) combine( &_signal.front(), &_kernel.front() );
      SLATEC_RFFTB1(_length,&_signal.front(),&_CH.front(),&_WA.front(),_IFAC);
      _isOpen = false;
   }
   return _signal;
}
