// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997 - 2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CubicHermiteInterpolant.cc --- Implementation of CubicHermiteInterpolants
//
//	DESCRIPTION:
//		Implementation of the CubicHermiteInterpolant class via the
//		SLATEC routines DPCHIM/DPCHFE/DPCHIA respectively. cfortran.h is used
//		to interface to SLATEC.
//
//	CHANGE HISTORY:
//		UL, 1999-09-08: 1.0.0: - first version
//		UL, 2000-02-01: 1.1.0: - added integral() - implementation via DPCHIA
//		UL, 2000-03-13: 1.2.0: - added copy constructor/constant constructor/
//								 assignment opertator + utility operators
//		UL, 2001-06-11: 1.3.0: - added supplement() method
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/slatec/cxx/CubicHermiteInterpolant.cc,v 1.13 2009/02/06 17:51:41 eojero Exp $
//

#include <functional>
#include <algorithm>
#include "slatec++.h"
#include "slatec.h"
#include "DataVector.h"
#include "errstr.h"
#include "Tracer.h"

using namespace std;
namespace Slatec {

template <> void CubicHermiteInterpolant<double>::free()
{
	PROC

	delete [] _x; _x = 0;
	delete [] _y; _y = 0;
	delete [] _deriv; _deriv = 0;

	ENDPROC
} // CubicHermiteInterpolant<double>::free()


template <> void CubicHermiteInterpolant<double>::initFromVectors(const vector<double> &x,	const vector<double> &y)
{
	PROC

	if (x.size() < 2 || y.size() < 2)
		errstr<<"Need at least two data points to define a cubic Hermite "
				"interpolant"<<Sas::error("InsufficentInput");

	if (x.size() != y.size())
		errstr<<"Number of abscissa values ("<<x.size()<<") is not equal to "
				"number of values ("<<y.size()<<")"
			  <<Sas::error("DataMismatch");

	_n = x.size();
	_x = new double[_n];
	_y = new double[_n];
	_deriv = new double[_n];

	for (unsigned n=0; n<_n; ++n) {
		_x[n] = x[n];
		_y[n] = _f*y[n] + _a;
	}

	_f = 1.;
	_a = 0.;

	ENDPROC
} // CubicHermiteInterpolant<double>::initfromVectors()


template <> void CubicHermiteInterpolant<double>::initSlatec()
{
	PROC

	int ierr = 0;
	SLATEC_DPCHIM(_n, _x, _y, _deriv, ierr);

	if (ierr) {
		if (ierr < 0) {
			switch (ierr) {
				case -3: errstr<<"The abcissas to define a cubic Hermite "
								 "interpolant are not strictly increasing"
							   <<Sas::error("InvalidInput");
						 break;
				default: // the other possible errors are caught above
						 errstr<<"This should not happen"
							   <<Sas::fatal("CubicHermiteInterpolant");
						 break;
			              }
		 } else {
			errstr<<ierr<<" switches in the direction of monotonicity were "
				  "detected during construction of cubic hermite inerpolant "
				  <<Sas::message(Msg::Lib, Msg::Verbose);
                        }
                 }
	ENDPROC
} // CubicHermiteInterpolant::initSlatec()


template <> CubicHermiteInterpolant<double>::CubicHermiteInterpolant() :
	Interpolant<double>(),
	_n(0), _x(0), _y(0), _deriv(0), _f(1.), _a(0.)
{
	PROC
	ENDPROC
} // CubicHermiteInterpolant<double>::CubicHermiteInterpolant()


template <> CubicHermiteInterpolant<double>::CubicHermiteInterpolant(double a) :
	Interpolant<double>(),
	_f(1.), _a(0.)
{
	PROC

	vector<double> x(2), y(2);
	x[0] = -1.e8;
	x[1] = 1.e8;
	y[0] = y[1] = a;
	_abscissa = x;

	initFromVectors(x, y);
	initSlatec();	

	ENDPROC
} // CubicHermiteInterpolant<double>::CubicHermiteInterpolant()


template <> CubicHermiteInterpolant<double>::CubicHermiteInterpolant(
								const vector<double> &x,
								const vector<double> &y)
	: Interpolant<double>(x), _f(1.), _a(0.)
{
	PROC

	initFromVectors(x, y);
	initSlatec();

	ENDPROC
} // CubicHermiteInterpolant<double>::CubicHermiteInterpolant()


template <> CubicHermiteInterpolant<double>::~CubicHermiteInterpolant()
{
	PROC

	free();

	ENDPROC
} // CubicHermiteInterpolant<double>::~CubicHermiteInterpolant()


template <> CubicHermiteInterpolant<double> &
	CubicHermiteInterpolant<double>::operator=(
			const CubicHermiteInterpolant<double> &rhs)
{
	PROC

	if (&rhs == this)
		ALTRETURNV(*this);

	Interpolant<double>::operator=(rhs);

	_n = rhs._n;
	_a = rhs._a;
	_f = rhs._f;
	delete [] _x; _x = new double[_n];
	delete [] _y; _y = new double[_n];
	delete [] _deriv; _deriv = new double[_n];

	memcpy(_x, rhs._x, _n * sizeof(double));
	memcpy(_y, rhs._y, _n * sizeof(double));
	memcpy(_deriv, rhs._deriv, _n * sizeof(double));

	RETURN *this;
} // CubicHermiteInterpolant<double>::operator=()


template <> CubicHermiteInterpolant<double>::CubicHermiteInterpolant(
			const CubicHermiteInterpolant<double> &r)
	: Interpolant<double>(r._abscissa)
{
	PROC

	_x = _y = _deriv = 0;
	this->operator=(r);
	ENDPROC
} // CubicHermiteInterpolant<double>::CubicHermiteInterpolant()



template <> CubicHermiteInterpolant<double> &
	CubicHermiteInterpolant<double>::operator*=(double f)
{
	PROC

	_f *= f;

	RETURN *this;
} // CubicHermiteInterpolant<double>::operator*=()


template <> CubicHermiteInterpolant<double> &
	CubicHermiteInterpolant<double>::operator*=(const vector<double> &v)
{
	PROC

	if (v.size() != _abscissa.size())
		errstr<<"Abscissa of CubicHermiteInterpolant has different size ("
			  <<_abscissa.size()<<") than vector factor ("<<v.size()<<")"
			  <<Sas::error("SizeMismatch");
	unsigned N = v.size();
	for (unsigned n=0; n<N; ++n)
		_y[n] = (_f * _y[n] + _a) * v[n];

	_f = 1.;
	_a = 0.;
	
	initSlatec();

	RETURN *this;
} // CubicHermiteInterpolant<double>::operator*=()


template <> CubicHermiteInterpolant<double> &
	CubicHermiteInterpolant<double>::operator+=(double a)
{
	PROC

	_a += a;

	RETURN *this;
} // CubicHermiteInterpolant<double>::operator+=()


template <> CubicHermiteInterpolant<double> &CubicHermiteInterpolant<double>::
	operator*=(const Interpolant<double> &f)
{
	PROC

	for (unsigned n=0; n<_n; ++n)
		_y[n] = (_f * _y[n] + _a) * f(_x[n]);
	_f = 1.;
	_a = 0.;

	initSlatec();

	RETURN *this;
} // CubicHermiteInterpolant<double>::operator*=()


template <> CubicHermiteInterpolant<double> &CubicHermiteInterpolant<double>::
	operator/=(const Interpolant<double> &f)
{
	PROC

	for (unsigned n=0; n<_n; ++n) {
		double denom = f(_x[n]);
		if (denom == 0.)
			denom = 1.e-16;
		_y[n] = (_f * _y[n] + _a)/denom;
	}

	_f = 1.;
	_a = 0.;

	initSlatec();

	RETURN *this;
} // CubicHermiteInterpolant<double>::operator/=()


template <> CubicHermiteInterpolant<double> &CubicHermiteInterpolant<double>::
	operator+=(const Interpolant<double> &a)
{
	PROC

	for (unsigned n=0; n<_n; ++n)
		_y[n] = (_f * _y[n] + _a) + a(_x[n]);

	_f = 1.;
	_a = 0.;

	initSlatec();

	RETURN *this;
} // CubicHermiteInterpolant<double>::operator*=()


template <> double CubicHermiteInterpolant<double>::operator()(double x) const
{
	PROC

	int ierr = 0, skip = 1;
	double value;

	SLATEC_DPCHFE(_n, _x, _y, _deriv, skip, 1, &x, &value, ierr);

	if (ierr > 0)
		errstr<<"Extrapolation was performed at "<<ierr<<" point(s) in order "
				"to evaluate cubic Hermite interpolant at x="<<x
			  <<Sas::message(Msg::Lib, Msg::Verbose);
	else if (ierr < 0)
		//	all other possible errors have been caught in the constructor
		errstr<<"This should not happen"
			  <<Sas::fatal("CubicHermiteInterpolant");

	RETURN value * _f + _a;
} // CubicHermiteInterpolant<double>::operator()


template <> CubicHermiteInterpolant<double> &CubicHermiteInterpolant<double>::supplement(
											const vector<double> &xsuppl,
											const vector<double> &ysuppl)
{
	PROC

	if (xsuppl.size() != ysuppl.size())
		errstr<<"Abscissa and ordinate data vectors have different size - "
				"cannot supplement cubic Hermite interpolant"
			  <<Sas::error("VectorSizeMismatch");

	if (xsuppl.size()) {
		vector<double> y(_y, _y+_n);
		CorrelatedVectors<double> corrVec(_abscissa, y);
		corrVec.supplement(xsuppl, ysuppl);

		//	re-initialize internals
		free();
		initFromVectors(corrVec.free(), corrVec.dependent());
		initSlatec();
	}

	RETURN *this;
} // CubicHermiteInterpolant<double>::supplement()


template <> CubicHermiteInterpolant<double> &CubicHermiteInterpolant<double>::replace(
											const vector<double> &xsuppl,
											const vector<double> &ysuppl)
{
	PROC

	if (xsuppl.size() != ysuppl.size())
		errstr<<"Abscissa and ordinate data vectors have different size - "
				"cannot replace cubic Hermite interpolant"
			  <<Sas::error("VectorSizeMismatch");

	if (!DataVector::sorted(xsuppl))
		errstr<<"Abscissa data vector is not sorted - cannot replace "
				"cubic Hermite interpolant"
			  <<Sas::error("VectorUnsorted");

	if (xsuppl.size()) {
		vector<double> y(_y, _y+_n), &x = _abscissa;
		typedef vector<double>::iterator Iter;
		Iter lb = lower_bound(x.begin(), x.end(), xsuppl.front()),
			 ub = lower_bound(x.begin(), x.end(), xsuppl.back());

		if (ub != x.end() && *ub == xsuppl.back())
			++ub;

		//	instantiate CubicSplineInterplant from xsuppl/ysuppl and
		//	evaluate it on intermediate points
		vector<double> xadd, yadd;
		if (lb != ub) {
			CubicHermiteInterpolant<double> inter(xsuppl, ysuppl);
			for (Iter i=lb; i!=ub; ++i) {
				xadd.push_back(*i);
				yadd.push_back(inter(*i));
			}
		}

		//	now merge the two vectors ...
		vector<double> xnew(xsuppl),
					   ynew(ysuppl);
		CorrelatedVectors<double> newVec(xnew, ynew);
		newVec.supplement(xadd, yadd);
	
		//	... and replace range in orignal vectors ...
		unsigned from = lb-x.begin(),
				 to = ub-x.begin();
		x.erase(lb, ub);
		x.insert(from+x.begin(), xnew.begin(), xnew.end());
		y.erase(y.begin()+from, y.begin()+to);
		y.insert(from+y.begin(), ynew.begin(), ynew.end());

		//	... and re-initialize internals
		free();
		initFromVectors(x, y);
		initSlatec();
	}

	RETURN *this;
} // CubicHermiteInterpolant<double>::replace()


template <> CubicHermiteInterpolant<double> &CubicHermiteInterpolant<double>::add(
											const vector<double> &xsuppl)
{
	PROC

	unsigned N = xsuppl.size();
	vector<double> ysuppl(N);
	for (unsigned n=0; n<N; ++n)
		ysuppl[n] = operator()(xsuppl[n]);

	RETURN supplement(xsuppl, ysuppl);
} // CubicHermiteInterpolant<double>::add()


template <> double CubicHermiteInterpolant<double>::integral(double x1, double x2) const
{
	PROC

	int	ierr = 0, skip = 1;
	double res = SLATEC_DPCHIA(_n, _x, _y, _deriv, skip, x1, x2, ierr);

	if (ierr)
		switch (ierr) {
			case 1:	errstr<<x1<<" is outside the interval ["<<_x[0]<<", "
						  <<_x[_n-1]<<"] on which the cubic Hermite "
							"interpolant is defined - intergration result "
							"might be afflicted with large error"
						  <<Sas::warning("PossiblyInaccurateResult");
					break;
			case 2:	errstr<<x2<<" is outside the interval ["<<_x[0]<<", "
						  <<_x[_n-1]<<"] on which the cubic Hermite "
							"interpolant is defined - intergration result "
							"might be afflicted with large error"
						  <<Sas::warning("PossiblyInaccurateResult");
					break;
			case 3: errstr<<x1<<" and "<<x2<<" are outside the interval ["
						  <<_x[0]<<", "<<_x[_n-1]<<"] on which the cubic "
							"Hermite interpolant is defined: ["
						  <<x1<<", "<<x2<<"] contains the entire data "
							"interval or the intervals do not intersect "
							"at all - integration result might be afflicted "
							"with large error."
						  <<Sas::warning("PossiblyInaccurateResult");
					break;
			default: errstr<<"Unknown numerical error "<<ierr<<" occured "
							 "in SLATEC routine DPCHIA"
						   <<Sas::error("NumericalError");
					break;
		}
							
	RETURN _f * res + _a * (x2-x1);
} // CubicHermiteInterpolant<double>::integral()


template <> void CubicHermiteInterpolant<double>::printOn(ostream &os) const
{
	PROC

	os<<"n   : "<<_n<<endl;
	os<<"f/a : "<<_f<<'/'<<_a<<endl;
	os<<"x    y    deriv"<<endl;
	for (unsigned n=0; n<_n; ++n)
		os<<_x[n]<<' '<<_y[n]<<' '<<_deriv[n]<<endl;

	ENDPROC
} // operator<<

} // end of namespace ?

template<class NumericType>
ostream &operator<<(ostream &os,
					const Slatec::CubicHermiteInterpolant<NumericType> &f)
{
	PROC

	f.printOn(os);
	RETURN os;
} // operator<<

template ostream &operator<<(ostream &os,
					const Slatec::CubicHermiteInterpolant<double> &f);


