// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997 - 2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CubicSplineInterpolant.h --- Implementation of Interpolant class as
//								  cubic Spline functions
//
//	DESCRIPTION:
//		A CubicSplineInterpolant object represents a monotone piecewise
//		cubic spline interpolant to given data.
//
//	CHANGE HISTORY:
//		UL, 1999-09-08: 1.0.0: - first version
//		UL, 2000-02-01: 1.1.0: - added integral()
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/slatec/cxx/CubicSplineInterpolant.h,v 1.8 2001/11/06 17:07:22 ulammers Exp $
//

#ifndef CUBIC_Spline_INTERPOLANT_H
#define CUBIC_Spline_INTERPOLANT_H

#include "Interpolant.h"

namespace Slatec {

template<class NumericType>
class CubicSplineInterpolant : public Interpolant<NumericType> {
public:
	CubicSplineInterpolant(const std::vector<NumericType> &x,
						   const std::vector<NumericType> &y);
		//
		//	PURPOSE:
		//		Instantiate a CubicSplineInterpolant object
		//
		//	ARGUMENTS:
		//		x	: vector of independent variable values
		//		y	: vector of dependent variable values corresponding to <X>
		//


	NumericType operator()(NumericType x) const;
		//
		//	PURPOSE:
		//		Evaluate interpolation function at a point <X>
		//


	const Interpolant<NumericType> &operator*=(NumericType);
	const Interpolant<NumericType> &operator*=
									(const Interpolant<NumericType> &);
	const Interpolant<NumericType> &operator+=
									(const Interpolant<NumericType> &);
	const Interpolant<NumericType> &operator+=(NumericType);
		//
		//	PURPOSE:
		//		Perform arithmetic on CubicHermiteInterpolant
		//


	Interpolant<NumericType> &supplement(const std::vector<NumericType> &xsupp,
									 const std::vector<NumericType> &ysupp);
		//
		//	PURPOSE:
		//		Supplement interpolated curve by additional	data pairs
		//		in <xsupp>/<ysupp> vectors
		//


	NumericType integral(NumericType x1, NumericType x2) const;
		//
		//	PURPOSE:
		//		Calculate integral of interpolant between <x1> and <x2>
		//


	~CubicSplineInterpolant();
		//
		//	PURPOSE:
		//		Destroy object instantiated with above constructor(s).
		//

private:
	unsigned	_n;
	NumericType	*_x, *_y, *_deriv;
	NumericType	_f, _a;

	void initSlatec();
	void initFromVectors(const std::vector<NumericType> &x,
						 const std::vector<NumericType> &y);
	void free();
};

}

#endif
