// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997 - 2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	LinearInterpolant.cc --- Implementation of LinearInterpolants
//
//	DESCRIPTION:
//		Implementation of the LinearInterpolant class.
//
//	CHANGE HISTORY:
//		UL, 2000-09-28: 1.0.0: - first version
//		UL, 2000-10-04: 1.1.0: - implemented integrate() method
//		UL, 2000-10-13: 1.2.0: - bug fix (made _startLi member variable)
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/slatec/cxx/LinearInterpolant.cc,v 1.8 2004/05/04 16:08:12 eojero Exp $
//

#include "slatec++.h"
#include "DataVector.h"
#include "slatec.h"
#include "errstr.h"
#include "Tracer.h"

namespace Slatec {

template<class T>
void LinearInterpolant<T>::validate() const
{
	PROC

	if (this->_abscissa.size() != _y.size() ||
		_y.size() != _n)
		errstr<<"The x/y data vectors have different sizes"
			  <<Sas::error("DifferentVectorSizes");

	if (_n < 2)
		errstr<<"Need at least two data points to define a LinearInterpolant"
			  <<Sas::error("TooLittleDataPoints");

	//
	//	check if abcissa vector is sorted
	//
	if (!DataVector::sorted(this->_abscissa))
		errstr<<"Abscissa value are not monotonically increasing which "
				"is a pre-requesite for a LinearInterpolant"
			  <<Sas::error("AbscissaNotSorted");

	ENDPROC
} // LinearInterpolant::validate()


template<class T>
LinearInterpolant<T>::LinearInterpolant() : _n(0), _startLi(0), _f(1.), _a(0.)
{
	PROC
	ENDPROC
} // LinearInterpolant<T>::LinearInterpolant()


template<class T>
LinearInterpolant<T>::LinearInterpolant(T a) :
	_startLi(0), _f(1.), _a(0.)
{
	PROC

	_n = 2;
	this->_abscissa.resize(_n);
	_y.resize(_n);
	
	this->_abscissa[0] = -1.e8;
	this->_abscissa[1] = 1.e8;
	_y[0] = _y[1] = a;

	validate();

	ENDPROC
} // LinearInterpolant<T>::LinearInterpolant()


template<class T>
LinearInterpolant<T>::LinearInterpolant(const std::vector<T> &x,
										const std::vector<T> &y)
	: Interpolant<double>(x), _startLi(0), _f(1), _a(0)
{
	PROC

	_y = y;
	_n = _y.size();
	validate();

	ENDPROC
} // LinearInterpolant<T>::LinearInterpolant()


template<class T>
LinearInterpolant<T> &LinearInterpolant<T>::operator=(
											const LinearInterpolant<T> &rhs)
{
	_startLi = 0;
	_n = rhs._n;
	_a = rhs._a;
	_f = rhs._f;
	_y = rhs._y;

	RETURN *this;
} // LinearInterpolant<T>::operator=()


template<class T>
LinearInterpolant<T>::LinearInterpolant(const LinearInterpolant<T> &r)
  : Slatec::Interpolant<T>()
{
	PROC

	this->operator=(r);

	ENDPROC
} // LinearInterpolant<T>::LinearInterpolant()


template<class T>
LinearInterpolant<T> &LinearInterpolant<T>::operator*=(T f)
{
	PROC

	_f *= f;

	RETURN *this;
} // LinearInterpolant<T>::operator*=()


template<class T>
LinearInterpolant<T> &LinearInterpolant<T>::operator*=(const std::vector<T> &v)
{
	PROC

	if (v.size() != _n)
		errstr<<"Abscissa of LinearInterpolant has different size ("
			  <<this->_abscissa.size()<<") than vector factor ("<<v.size()<<")"
			  <<Sas::error("SizeMismatch");
	unsigned N = v.size();
	for (unsigned n=0; n<N; ++n)
		_y[n] = y(n) * v[n];

	_f = 1.;
	_a = 0.;
	
	validate();

	RETURN *this;
} // LinearInterpolant<T>::operator*=()


template<class T>
LinearInterpolant<T> &LinearInterpolant<T>::operator+=(T a)
{
	PROC

	_a += a;

	RETURN *this;
} // LinearInterpolant<T>::operator+=()


template<class T>LinearInterpolant<T> &LinearInterpolant<T>::
	operator*=(const Interpolant<T> &f)
{
	PROC

	for (unsigned n=0; n<_n; ++n)
		_y[n] = y(n) * f(this->_abscissa[n]);
	_f = 1.;
	_a = 0.;

	validate();

	RETURN *this;
} // LinearInterpolant<T>::operator*=()


template<class T>
LinearInterpolant<T> &LinearInterpolant<T>::operator/=(const Interpolant<T> &f)
{
	PROC

	for (unsigned n=0; n<_n; ++n) {
		double denom = f(this->_abscissa[n]);
		if (denom == 0.)
			denom = 1.e-16;
		_y[n] = y(n)/denom;
	}

	_f = 1.;
	_a = 0.;

	validate();

	RETURN *this;
} // LinearInterpolant<T>::operator/=()


template<class T>
LinearInterpolant<T> &LinearInterpolant<T>::operator+=(const Interpolant<T> &a)
{
	PROC

	for (unsigned n=0; n<_n; ++n)
		_y[n] = y(n) + a(this->_abscissa[n]);

	_f = 1.;
	_a = 0.;

	validate();

	RETURN *this;
} // LinearInterpolant<T>::operator*=()


template<class T>
unsigned LinearInterpolant<T>::leftIndex(T x, unsigned startLi) const
{
	PROC

	//
	//	search index li such that
	//	this->_abscissa[li] <= x < this->_abscissa[li+1]
	//
	unsigned li;
	if (this->_abscissa[startLi] <= x &&
							(this->_abscissa[startLi+1] > x || startLi+1 == _n-1))
		li = startLi;
	else {
		li = 0;
		unsigned ri = _n-1,
		i = (li+ri)/2;
		while (li < ri-1) {
			if (x >= this->_abscissa[i])
				li = i;
			else if (x < this->_abscissa[i])
				ri = i;
			else
				break;
			i = (li+ri)/2;
		}
	}

	RETURN li;
} // LinearInterpolant<T>::leftIndex()
	

template<class T>
T LinearInterpolant<T>::operator()(T x) const
{
	PROC

	unsigned li = leftIndex(x, _startLi);
	if (x < this->_abscissa[li] || x > this->_abscissa[li+1])
		errstr<<"Performing extrapolation to evaluate LinearInterpolant at "
				"x="<<x<<Sas::message(Msg::Lib, Msg::Verbose);
	_startLi = li;

	RETURN y(li) +
			(y(li+1)-y(li))/(this->_abscissa[li+1]-this->_abscissa[li])*(x-this->_abscissa[li]);
} // LinearInterpolant<T>::operator()


template<class T>
T LinearInterpolant<T>::integral(T a, T b) const
{
	PROC

	if (a == b)
		ALTRETURNV(0);
	if (a > b)
		ALTRETURNV(-integral(b, a));

	double sum = 0.;
	unsigned li = leftIndex(a);
	if (a > this->_abscissa[li])
		++li;
	sum += .5*(operator()(a)+_y[li])*(this->_abscissa[li]-a);

	unsigned ri = leftIndex(b);
	if (b > this->_abscissa[ri+1])
		++ri;
	sum += .5*(operator()(b)+_y[ri])*(b-this->_abscissa[ri]);

	//	sum up areas of trapezoids in [a, b]
	for (unsigned i=li; i<ri; ++i)
		sum += .5*(_y[i]+_y[i+1])*(this->_abscissa[i+1]-this->_abscissa[i]);

	RETURN sum;
} // LinearInterpolant<T>::integral()

template<class T>
Interpolant<T> &LinearInterpolant<T>::supplement(const std::vector<T> &xsuppl,
												 const std::vector<T> &ysuppl)
{
	PROC

	CorrelatedVectors<T> corrVec(this->_abscissa, _y);
	corrVec.supplement(xsuppl, ysuppl);
	validate();

	RETURN *this;
} // LinearInterpolant<T>::supplement()


template<class T>
void LinearInterpolant<T>::printOn(std::ostream &os) const
{
	PROC

	os<<"n   : "<<_n<<std::endl;
	os<<"f/a : "<<_f<<'/'<<_a<<std::endl;
	os<<"x    y    "<<std::endl;
	for (unsigned n=0; n<_n; ++n)
		os<<this->_abscissa[n]<<' '<<_y[n]<<' '<<std::endl;

	ENDPROC
} // operator<<


template<class T>
LinearInterpolant<T>::~LinearInterpolant()
{
	PROC
	ENDPROC
} // LinearInterpolant<T>::~LinearInterpolant()


}

template<class NumericType>
std::ostream &operator<<(std::ostream &os,
					const Slatec::LinearInterpolant<NumericType> &f)
{
	PROC

	f.printOn(os);
	RETURN os;
} // operator<<

//	explicit instantiation
template std::ostream &operator<<(std::ostream &os,
							 const Slatec::LinearInterpolant<double> &f);

template class Slatec::LinearInterpolant<double>;
