// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// Giuseppe Vacanti, September 1999
//
// Remarks:
//
//  - SLATEC routines are available in single precision only. We like
//  a double precision interface, so we cast up/down as required.
//
//  - Should provide an object interface.
//  

#include <Tracer.h>
#include "slatec.h"
#include <errstr.h>
#include <vector>
#include "Convolver.h"

namespace Slatec {

  std::vector<double> forwardFft(const std::vector<double> & x)
    {
      PROC
      unsigned n = x.size();

      float * ch = new float[n];
      float * wa = new float[n];
      int * ifac = new int[15];
      SLATEC_RFFTI1(n,wa,ifac);      // initialize work arrays

      float * s = new float[n];
      for(unsigned i=0; i<n; i++)
	s[i] = static_cast<float>(x[i]);
      // s is overwritten by the following call
      SLATEC_RFFTF1(n,s,ch,wa,ifac);
      std::vector<double> c(n);
      for(unsigned i=0; i<n; i++)
	c[i] = static_cast<double>(s[i]);
      delete[] ch;
      delete[] wa;
      delete[] ifac;
      delete[] s;
      RETURN c;
    }

  std::vector<double> backwardFft(const std::vector<double> & x)
    {
      PROC
      unsigned n = x.size();
      
      float * ch = new float[n];
      float * wa = new float[n];
      int * ifac = new int[15];
      SLATEC_RFFTI1(n,wa,ifac);      // initialize work arrays

      float * s = new float[n];
      for(unsigned i=0; i<n; i++)
	s[i] = static_cast<float>(x[i]);
      // s is overwritten by the following call
      SLATEC_RFFTB1(n,s,ch,wa,ifac);
      std::vector<double> c(n);
      for(unsigned i=0; i<n; i++)
	c[i] = static_cast<double>(s[i]);
      delete[] ch;
      delete[] wa;
      delete[] ifac;
      delete[] s;
      RETURN c;
    }

  // convolve by Doug Reynolds

  std::vector<double> convolve( const std::vector<double> &signal,
			   const std::vector<double> &response )
    {
      PROC
	if ( signal.size() != response.size() )
	  errstr << "Signal and response vector have differing sizes: "
		 << signal.size() << " " << response.size()
		 << Sas::error("Slatec::convolve: size mismatch");
      
      unsigned i, n = signal.size();
      float *s = new float [n], *r = new float [n];
      for( i = 0; i < n; ++i ) {
	s[i] = signal[i];
	r[i] = response[i];
      }
      Convolver conv(n);
      conv.load(s);
      conv.load(r);
      const std::vector<float> &c = conv();
      std::vector<double> convolution(n);
      for( i = 0; i < n; ++i ) convolution[i] = c[i];
      delete [] r; delete [] s;
      RETURN convolution;
    }
  
};
  
