// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//---------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2009 European Space Agency
//
//---------------------------------------------------------------------------
//
//	slatec.h: --- C/C++-Wrappers to subset of FORTRAN SLATEC routines ---
//		  Fortran routines from other software collections are also wrapped here.
//
//	DESCRIPTION:
//      This header files provides macros to create C-wrapper functions for
//      a subset of numerical routines from the public-domain FORTRAN library
//      SLATEC. The general C-Fortran interface `cfortran.h' is used.
//      For a description of the functionality and parameters of the various
//      functions, see file `slatec.f'.
//
//	MACROS:
//		TOMS_RGBI3P   --- two dimensional interpolation with rect. grid points
//      SLATEC_DPCHIM --- compute derivatives of cubic Hermite interpolant
//      SLATEC_DPCHFE --- evaluate cubic Hermite fun. using outout from DPCHIM
//		SLATEC_DPCHIA --- integrate cubic Hermite interpolant
//		SLATEC_DPCHSP --- get derivatives of cubic Hermite spline interpolant
//      SLATEC_DPOLFT --- least-square polynomial fit of discrete data
//      SLATEC_DP1VLU --- evaluate polynomial computed by DPOLFT
//		SLATEC_DERF	  --- error function
//		SLATEC_DGAUS8 --- integration of analytical function
//
//	CHANGE HISTORY:
//		UL, 1998-07-20: - initial version
//		UL, 1999-09-09: - added SLATEC_DPCHSP
//						- added SLATEC_DERF (GV)
//						- added '___dum' hack below to prevent 'defined but
//						  not used' warning from egcs/gcc
//		GV, 1999-09-13: - added DERFC, RFFTI1, RFFTB1, RFFTF1
//		GV, 2000-01-07: - added IDBVIP
//		UL, 2000-01-25: - added TOME_RGBI3P
//		UL, 2000-02-01: - added SLATEC_DPCHIA
//		UL, 2000-03-27: - added SLATEC_DGAUS8
//		UL, 2002-07-15: - made ___dum() standard compliant
//	   JBa, 2002-07-17: - made ___dum() standard compliant(, really)
//
//	AUTHOR:
//		ULammers, Mon Jul 20 15:38:18 MET DST 1998
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/slatec/cxx/slatec.h,v 1.12 2002/07/17 12:09:49 jbakker Exp $

#ifndef SLATEC_H
#define SLATEC_H

#include "cfortran.h"
#include "common.h"

//#define SLATEC_AS_MODULE	//	uncomment if SLATEC is a module

#if !defined(AS_MODULE)
	#if defined(F90MODPROC)
		#undef F90MODPROC
		#define F90MODPROC(a, b)	CXXFUNC(b, b)
	#else
		#define F90MODPROC(a, b)	CXXFUNC(b, b)
	#endif
#endif


//
//	XERMSG: write error message to stdout
//
PROTOCCALLSFSUB5(F90MODPROC(slatec, xermsg), xmermsg,					\
				STRING, STRING, STRING, INT, INT)
#define SLATEC_XERMSG(S1, S2, S3, I1, I2)								\
	CCALLSFSUB5(F90MODPROC(slatec, xermsg), xmermsg,					\
				STRING, STRING, STRING, INT, INT,						\
				S1, S2, S3, I1, I2)

//
//	DPCHIM:	calculate derivatives needed to determine a monotone piecewise
//          cubic Hermite interpolant to given data.
//			Note: Original parameter INCFD (cmp. slatec.f) has been omitted
//				  since we only need to do 1-dimensional interpolations
//
PROTOCCALLSFSUB6(F90MODPROC(slatec, dpchim), dpchim,                    \
                INT, DOUBLEV, DOUBLEV, DOUBLEV, INT, PINT)
#define SLATEC_DPCHIM(N, X, F, D, IERR)									\
    CCALLSFSUB6(F90MODPROC(slatec, dpchim), dpchim,                     \
                INT, DOUBLEV, DOUBLEV, DOUBLEV, INT, PINT,				\
                N, X, F, D, 1, IERR)

//
//	DPCHSP:	Set derivatives needed to determine the Hermite represen-
//			tation of the cubic spline interpolant to given data, with
//			specified boundary conditions.
//			Note: Original parameter INCFD (cmp. slatec.f) has been omitted
//				  since we only need to do 1-dimensional interpolations
//
PROTOCCALLSFSUB10(F90MODPROC(slatec, dpchsp), dpchsp,                   \
                INTV, DOUBLEV, INT, DOUBLEV, DOUBLEV, DOUBLEV, INT,		\
				DOUBLEV, INT, PINT)
#define SLATEC_DPCHSP(IC, VC, N, X, F, D, WK, NWK, IERR)				\
    CCALLSFSUB10(F90MODPROC(slatec, dpchsp), dpchsp,                    \
                INTV, DOUBLEV, INT, DOUBLEV, DOUBLEV, DOUBLEV, INT,		\
				DOUBLEV, INT, PINT,										\
                IC, VC, N, X, F, D, 1, WK, NWK, IERR)

//
//  DPCHFE: Evaluate a piecewise cubic Hermite function at an array of
//          points using output from DPCHIM or DPCHSP
//			Note: Original parameter INCFD (cmp. slatec.f) has been omitted
//				  since we only need to do 1-dimensional interpolations
//
PROTOCCALLSFSUB10(F90MODPROC(slatec, dpchfe), dpchfe,                   \
                 INT, DOUBLEV, DOUBLEV, DOUBLEV, INT, PLOGICAL, INT,	\
                 DOUBLEV, DOUBLEV, PINT)
#define SLATEC_DPCHFE(N, X, F, D, SKIP, NE, XE, FE, IERR)				\
    CCALLSFSUB10(F90MODPROC(slatec, dpchfe), dpchfe,                    \
                 INT, DOUBLEV, DOUBLEV, DOUBLEV, INT, PLOGICAL, INT,	\
                 DOUBLEV, DOUBLEV, PINT,                                \
                 N, X, F, D, 1, SKIP, NE, XE, FE, IERR)


//
//	DPCHIA: Evaluate the definite integral of a piecewise cubic
//			Hermite function over an arbitrary interval.
//			Note: Original parameter INCFD (cmp. slatec.f) has been omitted
//				  since we only deal with 1-dimensional interpolants
//
PROTOCCALLSFFUN9(DOUBLE, F90MODPROC(slatec, dpchia), dpchia,			\
				 INT, DOUBLEV, DOUBLEV, DOUBLEV, INT, PLOGICAL,			\
				 DOUBLE, DOUBLE, PINT)
#define SLATEC_DPCHIA(N, X, F, D, SKIP, A, B, IERR)						\
	CCALLSFFUN9(F90MODPROC(slatec, dpchia), dpchia,						\
				INT, DOUBLEV, DOUBLEV, DOUBLEV, INT, PLOGICAL,			\
				DOUBLE, DOUBLE, PINT,									\
				N, X, F, D, 1, SKIP, A, B, IERR)


//
//  DPOLFT: Fit discrete data in a least squares sense by polynomials
//          in one variable.
//
PROTOCCALLSFSUB10(F90MODPROC(slatec, dpolft), dpolft,                   \
                 INT, DOUBLEV, DOUBLEV, DOUBLEV, INT, PINT, PDOUBLE,    \
                 DOUBLEV, PINT, DOUBLEV)
#define SLATEC_DPOLFT(N, X, Y, W, MAXDEG, NDEG, EPS, R, IERR, A)        \
    CCALLSFSUB10(F90MODPROC(slatec, dpolft), dpolft,                    \
                 INT, DOUBLEV, DOUBLEV, DOUBLEV, INT, PINT, PDOUBLE,    \
                 DOUBLEV, PINT, DOUBLEV,                                \
                 N, X, Y, W, MAXDEG, NDEG, EPS, R, IERR, A)

//
//  DP1VLU: Use the coefficients generated by DPOLFT to evaluate the
//          polynomial fit at a specified point
//
PROTOCCALLSFSUB6(F90MODPROC(slatec, dp1vlu), dp1vlu,                    \
                INT, INT, DOUBLE, PDOUBLE, DOUBLEV, DOUBLEV)
#define SLATEC_DP1VLU(L, NDER, X, YFIT, YP, A)                          \
    CCALLSFSUB6(F90MODPROC(slatec, dp1vlu), dp1vlu,                     \
                INT, INT, DOUBLE, PDOUBLE, DOUBLEV, DOUBLEV,            \
                L, NDER, X, YFIT, YP, A)

//
//	DNLS1E: An easy-to-use code which minimizes the sum of the squares
//			of M nonlinear functions in N variables by a modification
//			of the Levenberg-Marquardt algorithm.
//
PROTOCCALLSFSUB12(F90MODPROC(slatec, dnls1e), dnls1e,                   \
				 ROUTINE, INT, INT, INT, DOUBLEV, DOUBLEV, DOUBLE,      \
				 INT, PINT, INTV, DOUBLEV, INT)
#define SLATEC_DNLS1E(FCN, IOPT, M, N, X, FVEC, TOL, NPRINT, INFO,		\
					  IW, WA, LWA)                                      \
    CCALLSFSUB12(F90MODPROC(slatec, dnls1e), dnls1e,                    \
				 ROUTINE, INT, INT, INT, DOUBLEV, DOUBLEV, DOUBLE,      \
				 INT, PINT, INTV, DOUBLEV, INT,                         \
                 FCN, IOPT, M, N, X, FVEC, TOL, NPRINT, INFO,           \
                 IW, WA, LWA)

  //
  //    RFFTF1: Compute the forward transform of a real, periodic sequence.
  //

PROTOCCALLSFSUB5(F90MODPROC(slatec, rfftf1), rfftf1, \
		INT, FLOATV, FLOATV, FLOATV, INTV)
#define SLATEC_RFFTF1(N, C, CH, WA, IFAC) \
	CCALLSFSUB5(F90MODPROC(slatec, rfftf1), rfftf1, \
	INT, FLOATV, FLOATV, FLOATV, INTV, \
	N, C, CH, WA, IFAC)

//	
//	RFFTB1: Compute the backward fast Fourier transform of a real
//			coefficient array.
//	
PROTOCCALLSFSUB5(F90MODPROC(slatec, rfftb1), rfftb1, \
		INT, FLOATV, FLOATV, FLOATV, INTV)
#define SLATEC_RFFTB1(N, C, CH, WA, IFAC) \
	CCALLSFSUB5(F90MODPROC(slatec, rfftb1), rfftb1, \
	INT, FLOATV, FLOATV, FLOATV, INTV, \
	N, C, CH, WA, IFAC)

//
//	RFFTI1: Initialize a real and an integer work array for RFFTF1 and
//			RFFTB1.
//
PROTOCCALLSFSUB3(F90MODPROC(slatec,rffti1), rffti1, \
		 INT, FLOATV, INTV)
#define SLATEC_RFFTI1(N, WA, IFAC) \
	CCALLSFSUB3(F90MODPROC(slatec,rffti1), rffti1, \
	INT, FLOATV, INTV, \
	N, WA, IFAC)

//
//    CFFTF1: Compute the forward transform of a complex, periodic sequence.
//
PROTOCCALLSFSUB5(F90MODPROC(slatec, cfftf1), cfftf1, \
		INT, FLOATV, FLOATV, FLOATV, INTV)
#define SLATEC_CFFTF1(N, C, CH, WA, IFAC) \
	CCALLSFSUB5(F90MODPROC(slatec, cfftf1), cfftf1, \
	INT, FLOATV, FLOATV, FLOATV, INTV, \
	N, C, CH, WA, IFAC)

//	
//		CFFTB1: Compute the unnormalized inverse of CFFTF1.
//	

PROTOCCALLSFSUB5(F90MODPROC(slatec, cfftb1), cfftb1, \
		INT, FLOATV, FLOATV, FLOATV, INTV)
#define SLATEC_CFFTB1(N, C, CH, WA, IFAC) \
	CCALLSFSUB5(F90MODPROC(slatec, cfftb1), cfftb1, \
	INT, FLOATV, FLOATV, FLOATV, INTV, \
	N, C, CH, WA, IFAC)

//
// 	CFFTI1: Initialize a real and an integer work array for CFFTF1 and CFFTB1.
//
PROTOCCALLSFSUB3(F90MODPROC(slatec,cffti1), cffti1, \
		 INT, FLOATV, INTV)
#define SLATEC_CFFTI1(N, WA, IFAC) \
	CCALLSFSUB3(F90MODPROC(slatec,cffti1), cffti1, \
	INT, FLOATV, INTV, \
	N, WA, IFAC)


/*  	 //	TOMS526 seems not to work -- January 2000 */
/*       // */
/*       // IDBVIP: TOMS526 -- bivariate interpolation with scattered (x,y) */
/*       // */
/*  PROTOCCALLSFSUB12(FMODPROC(slatec, idbvip), idbvip, \ */
/*  		  INT, INT, INT, FLOATV, FLOATV, FLOATV, INT, \ */
/*  		  FLOATV, FLOATV, FLOATV, INTV, FLOATV) */
/*  #define SLATEC_IDBVIP(MD, NCP, NDP, XD, YD, ZD, NIP, XI, YI, ZI, IWK, WK) \ */
/*  	CCALLSFSUB12(FMODPROC(slatec, idbvip), idbvip, \ */
/*  	INT, INT, INT, FLOATV, FLOATV, FLOATV, INT, \ */
/*  	FLOATV, FLOATV, FLOATV, INTV, FLOATV, \ */
/*  	MD, NCP, NDP, XD, YD, ZD, NIP, XI, YI, ZI, IWK, WK) */

/*       // */
/*       // ISDFFT: TOMS526 -- smooth surface fitting with scattered (x,y) */
/*       // */

/*  PROTOCCALLSFSUB13(FMODPROC(slatec, idsfft), idsfft, \ */
/*  		  INT, INT, INT, FLOATV, FLOATV, FLOATV, INT, \ */
/*  		  INT, FLOATV, FLOATV, FLOATV, INTV, FLOATV) */
/*  #define SLATEC_IDSFFT(MD, NCP, NDP, XD, YD, ZD, NXI, NYI, XI, YI, ZI, IWK, WK) \ */
/*  	CCALLSFSUB13(FMODPROC(slatec, idsfft), idsfft, \ */
/*  	INT, INT, INT, FLOATV, FLOATV, FLOATV, INT, \ */
/*  	INT, FLOATV, FLOATV, FLOATV, INTV, FLOATV, \ */
/*  	MD, NCP, NDP, XD, YD, ZD, NXI, NYI, XI, YI, ZI, IWK, WK) */


//
// TOMS660
//
PROTOCCALLSFSUB17(F90MODPROC(slatec, qshep2), qshep2, \
				  INT, FLOATV, FLOATV, FLOATV, INT, INT, INT, \
				  INTV, INTV, PFLOAT, PFLOAT, PFLOAT, PFLOAT, \
				  PFLOAT, FLOATV, FLOATV, PINT)
#define TOMS_QSHEP2(N,X,Y,F,NQ,NW,NR,LCELL,LNEXT,XMIN,YMIN,DX,DY,RMAX,RSQ,\
					  A,IER) \
		CCALLSFSUB17(F90MODPROC(slatec, qshep2), qshep2, \
				  	INT, FLOATV, FLOATV, FLOATV, INT, INT, INT, \
				  	INTV, INTV, PFLOAT, PFLOAT, PFLOAT, PFLOAT, \
				  	PFLOAT, FLOATV, FLOATV, PINT, \
				  	N,X,Y,F,NQ,NW,NR,LCELL,LNEXT,XMIN,YMIN,DX,DY,RMAX,RSQ,\
					A,IER)

PROTOCCALLSFSUB17(F90MODPROC(slatec, wrap_qs2val), wrap_qs2val, \
				  FLOAT, FLOAT, INT, FLOATV, FLOATV, FLOATV, \
				  INT, INTV, INTV, FLOAT, FLOAT, FLOAT, FLOAT, \
				  FLOAT, FLOATV, FLOATV, PFLOAT)
#define TOMS_WRAP_QS2VAL(PX,PY,N,X,Y,F,NR,LCELL,LNEXT,XMIN,YMIN,DX,DY,RMAX,RSQ,A,PZ) \
		CCALLSFSUB17(F90MODPROC(slatec, wrap_qs2val), wrap_qs2val, \
					FLOAT, FLOAT, INT, FLOATV, FLOATV, FLOATV, \
					INT, INTV, INTV, FLOAT, FLOAT, FLOAT, FLOAT, \
					FLOAT, FLOATV, FLOATV, PFLOAT, \
					PX,PY,N,X,Y,F,NR,LCELL,LNEXT,XMIN,YMIN,DX,DY,RMAX,RSQ,A,PZ)


//
// TOMS760: RGBI3P - 2D-interpolation based on rectangular grid points
//
PROTOCCALLSFSUB12(F90MODPROC(slatec, rgbi3p), rgbi3p, \
                  INT, INT, INT, FLOATV, FLOATV, FLOATV, INT, \
                  FLOATV, FLOATV, FLOATV, PINT, FLOATV)
#define TOMS_RGBI3P(MD,NXD,NYD,XD,YD,ZD,NIP,XI,YI, ZI,IER,WK) \
		CCALLSFSUB12(F90MODPROC(slatec, rgbi3p), rgbi3p, \
					INT, INT, INT, FLOATV, FLOATV, FLOATV, INT, \
					FLOATV, FLOATV, FLOATV, PINT, FLOATV, \
					MD,NXD,NYD,XD,YD,ZD,NIP,XI,YI,ZI,IER,WK)


//
//	DGAUS8: Integrate a real function of one variable over a finite
//			interval using an adaptive 8-point Legendre-Gauss
//			algorithm.
//	 
PROTOCCALLSFFUN6(DOUBLE, F90MODPROC(slatec, dgaus8), dgaus8, \
				 ROUTINE, DOUBLE, DOUBLE, DOUBLE, PDOUBLE, PINT)
#define SLATEC_DGAUS8(FUN, A, B, ERR, ANS, IERR) \
	CCALLSFFUN6(F90MODPROC(slatec, dgaus8), dgaus8, \
				ROUTINE, DOUBLE, DOUBLE, DOUBLE, PDOUBLE, PINT, \
				FUN, A, B, ERR, ANS, IERR)

//
//	DRAND: Random number generatr returning uniformally distributed number
//		   in interval [0.0, 1.0]
//
//PROTOCCALLSFFUN1(DOUBLE, DRAND, drand, DOUBLE)
//#define SLATEC_DRAND(R)
//	CCALLSFFUN1(DRAND, drand, DOUBLE, R)

//
// DERF: Compute the double precision error function
//	 
PROTOCCALLSFFUN1(DOUBLE, F90MODPROC(slatec, derf), derf, DOUBLE)
#define SLATEC_DERF(X) CCALLSFFUN1(F90MODPROC(slatec, derf), derf, DOUBLE, X)

//
// DERFC: Compute the double precision complementary error function
//	 
PROTOCCALLSFFUN1(DOUBLE, F90MODPROC(slatec, derfc), derfc, DOUBLE)
#define SLATEC_DERFC(X) CCALLSFFUN1(F90MODPROC(slatec, derfc), derfc, DOUBLE, X)

typedef void (*VFunvPtr)(void (*)());
static void ___dum(VFunvPtr) {
	___dum(reinterpret_cast<VFunvPtr>(___dum));
	___dum(reinterpret_cast<VFunvPtr>(_(CFFUN(derf),_)));
	___dum(reinterpret_cast<VFunvPtr>(_(CFFUN(derfc),_)));
	___dum(reinterpret_cast<VFunvPtr>(_(CFFUN(dpchia),_)));
	___dum(reinterpret_cast<VFunvPtr>(_(CFFUN(dgaus8),_)));
	// add here all interfaced SLATEC functions in the same manner;
	// this is to prevent a 'defined but not used warning' in egcs/gcc
}

#endif
